/*
 * $Id: //depot/prod/newport_news/main/modules/nn-codegen/src/main/java/com/netapp/common/generator/NetappClassGenerator.java#8 $
 *
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.generator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Class to handle building a string representing a Java class.
 * Includes support for things like maintaining indent, tracking imports,
 * simple Java string operations, etc.
 *
 * Generators for specific output classes generally extend this class to construct
 * the {@link #classBuffer} as required.
 *
 * @author taylor
 * @author sunday
 * @version $Revision: #8 $ $Date: 2012/11/13 $
 */
public class NetappClassGenerator {
    /**
     * Look for html &lt; &amp; &gt; tags
     * Group 1: Entire &lt;tag&gt; or &lt;/tag&gt; or \\
     * Group 2: Tag name or null
     * TODO: This needs unit tests to make sure it doesn't break
     */
    private static final Pattern DESC_SCRUB_PATTERN =
        Pattern.compile("(</?(\\w+)(?:\\s+[\\w\\_]+=\"(?:[^\"]|\\\\\")*\")*>|<|>|\\\\)");

    /**
     * List of allowed HTML tags. Tags in a comment that are not in this
     * list will be munged to html entities.
     */
    private static Set<String> allowedHtmlTags;;
    static {
        allowedHtmlTags = new HashSet<String>();
        allowedHtmlTags.add("b");
        allowedHtmlTags.add("br");
        allowedHtmlTags.add("font");
        allowedHtmlTags.add("i");
        allowedHtmlTags.add("p");
        allowedHtmlTags.add("dd");
        allowedHtmlTags.add("dl");
        allowedHtmlTags.add("dt");
        allowedHtmlTags.add("hr");
        allowedHtmlTags.add("i");
        allowedHtmlTags.add("li");
        allowedHtmlTags.add("ol");
        allowedHtmlTags.add("p");
        allowedHtmlTags.add("samp");
        allowedHtmlTags.add("table");
        allowedHtmlTags.add("tr");
        allowedHtmlTags.add("th");
        allowedHtmlTags.add("td");
        allowedHtmlTags.add("ul");
        allowedHtmlTags.add("var");

    }
    private CodeFormatter formatter;
    private StringBuffer classBuffer = new StringBuffer();
    private StringBuffer currentIndent = new StringBuffer();
    private String packageName;
    private List<String> packageComments;

    /**
     * A list of imports that we need for the generated code.
     */
    private Set<String> imports = new TreeSet<String>();

    /**
     * Create a new class generator inside the supplied package.
     * @param packageName Name of the package the generated code will reside in.
     */
    public NetappClassGenerator(String packageName) {
        this(packageName, new NetappCodeFormatter());
    }

    /**
     * Create a new class generator inside the supplied package with the
     * specified format.
     * @param packageName Name of the package the generated code will reside in.
     * @param formatter CodeFormatter to use to format the resulting class file.
     */
    public NetappClassGenerator(String packageName, CodeFormatter formatter) {
        super();
        this.packageName = packageName;
        this.formatter = formatter;
    }
    /**
     * Open the target file and write out the code that we've created.
     * @param dirName Directory to write the current class into
     * @param fileName File to write the current class into
     */
    @edu.umd.cs.findbugs.annotations.SuppressWarnings({"RV_RETURN_VALUE_IGNORED_BAD_PRACTICE", "DM_DEFAULT_ENCODING"})
    public final void writeCode(String dirName, String fileName) throws IOException {

        File dir = new File(dirName);
        dir.mkdirs();

        Writer fileWriter = new FileWriter(new File(dir, fileName));
        writeCode(fileWriter);
        fileWriter.close();
    }

    /**
     * Write out the code we've created to the specified FileWriter.
     * @param fileWriter Writer to write the current
     */
    public void writeCode(Writer fileWriter) throws IOException {
        fileWriter.write(formatter.fileHeader() + formatter.newLine());

        //
        if (packageComments != null && !packageComments.isEmpty()) {
            fileWriter.write("/**" + formatter.newLine());
            for (String comment : packageComments) {
                fileWriter.write(" * " + scrubJavadoc(comment) + formatter.newLine());
            }
            fileWriter.write(" */" + formatter.newLine());
        }

        fileWriter.write("package " + this.packageName + ";" + formatter.newLine());
        fileWriter.write(formatter.newLine());

        // Imports
        // We don't want to deal with ordering rules in generated code
        fileWriter.write("//CHECKSTYLE:OFF ImportOrder" + formatter.newLine());
        fileWriter.write("//CHECKSTYLE:OFF AvoidStarImport" + formatter.newLine());
        for (String str : imports) {
            fileWriter.write("import " + str + ";" + formatter.newLine());
        }
        fileWriter.write("//CHECKSTYLE:ON ImportOrder" + formatter.newLine());
        fileWriter.write("//CHECKSTYLE:ON AvoidStarImport" + formatter.newLine());

        // The class
        fileWriter.write(formatter.newLine());
        fileWriter.write(this.classBuffer.toString());
        fileWriter.flush();
    }

    /**
     * Get the string buffer containing the current class file.
     * The class buffer does not include headers and import declarations.
     * @return contents of the current class buffer
     */
    public StringBuffer getClassBuffer() {
        return classBuffer;
    }

    /**
     * Add a line of text to the class buffer at the current indent level.
     * @param line Line of text to add
     */
    public void addLine(String line, Object... args) {
        classBuffer.append(this.currentIndent);
        classBuffer.append(args.length > 0 ? String.format(line, args) : line);
        classBuffer.append(formatter.newLine());
    }

    public void addPackageJavadoc(String comment) {
        if (packageComments == null) {
            packageComments = new ArrayList<String>();
        }
        packageComments.add(comment);
    }

    /**
     * Add a blank line of text to the class buffer.
     *
     */
    public void addLine() {
        classBuffer.append(formatter.newLine());
    }

    /**
     * Add a GUID comment to the class buffer
     * @param guid
     */
    protected void addGuidLine(String guid) {
        addLine("// " + formatter.guid() + " " + guid);
    }

    protected String scrubJavadoc(String comment) {
        StringBuffer buffer = new StringBuffer();

        // This is rough, but we want to filter out impropper html-looking text
        // as well as escape characters
        Matcher m = DESC_SCRUB_PATTERN.matcher((comment != null) ? comment : "");

        while (m.find()) {
            if (m.group(2) != null
                    && allowedHtmlTags.contains(m.group(2).toLowerCase())) {
                m.appendReplacement(buffer, m.group(1).replaceAll("\\$", "\\\\\\$"));
            } else {
                // It takes 16 backslashes to result in two in the output.
                // Burt 371700: We have to escape '$' again before passing it as a value
                // to replaceAll.
                String rep = m.group(1)
                    .replaceAll("<", "&lt;")
                    .replaceAll(">", "&gt;")
                    .replaceAll("\\\\", "\\\\\\\\\\\\\\\\")
                    .replaceAll("\\$", "\\\\\\$");
                m.appendReplacement(buffer, rep);
            }
        }
        m.appendTail(buffer);
        return buffer.toString();
    }

    public void addJavadoc(List<String> comments) {
        addLine("/**");
        for (String line : comments) {
            String scrubbed = scrubJavadoc(line);
            for (String nl : scrubbed.split("[\\r\\n]+")) {
                addLine(" * " + nl);
            }
        }
        addLine(" */");
    }

    /**
     * Add a one line javadoc comment at the current indent level.
     * @param comment Javadoc comment to add.
     */
    public void addJavadocLine(String comment) {
        addJavadoc(Collections.singletonList(comment));
    }

    /**
     * Add unformatted text.
     *
     * @param text The text to add
     */
    public void addText(String text) {
        this.classBuffer.append(text);
    }

    /**
     * Add the current indent level to the class buffer
     *
     */
    public void addIndent() {
        this.classBuffer.append(this.currentIndent);
    }

    /**
     * Reset the current indent level to 0
     *
     */
    public void zeroIndent() {
        this.currentIndent.setLength(0);
    }

    /**
     * Increase the current indent level by one tab
     */
    public void increaseIndent() {
        this.currentIndent.append(formatter.tab());
    }

    /**
     * Increase the current indent level by one tab
     */
    public void indent() {
        this.increaseIndent();
    }

    /**
     * Decrease the current indent level by one tab
     */
    public void decreaseIndent() {
        try {
            this.currentIndent.delete(this.currentIndent.length() - formatter.tab().length(),
                this.currentIndent.length());
        } catch (StringIndexOutOfBoundsException e) {
            System.out.println("Too many decrease indents called.");
        }
    }

    /**
     * Decrease the current indent level by one tab
     */
    public void outdent() {
        this.decreaseIndent();
    }

    /**
     * Add the class to the import list.
     * Don't add "java.lang.*" types and types from our current package.
     * @param fullTypeName
     */
    public void addImport(String fullTypeName) {

        int index = fullTypeName.lastIndexOf('.');
        if (index != -1) {
            String importPackageName = fullTypeName.substring(0, index);

            // If the class is not in this package and not in java.lang, add it to our imports.
            if (!importPackageName.equals(this.packageName) && !importPackageName.equals("java.lang")) {
                imports.add(fullTypeName);
            }
        }
    }

    /**
     * Given a field name or package name, convert it to a class name by making the first letter upper case.
     *
     * @param fieldName
     * @return the class style name of fieldName
     */
    protected String toClassName(String fieldName) {
        return fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);
    }

    /**
     * Given a class name, create a variable name for an instance of that class by making the first letter lower case.
     *
     * @param className
     * @return the field style name of className
     */
    protected String toVariableName(String className) {
        return className.substring(0, 1).toLowerCase() + className.substring(1);
    }


    /**
     * Take a name with underscores and remove the underscores and capitalize the letters
     * after the underscores.
     * @param name
     * @return The java style name of name
     */
    protected String toJavaName(String name) {

        for (int i = 0; i < name.length(); i++) {
            if (name.charAt(i) == '_') {
                name = name.substring(0, i) + name.substring(i + 1, i + 2).toUpperCase() + name.substring(i + 2);
            }
        }
        return name;
    }

    /**
     * Given a full package name (e.g., com.netapp.dlm.repository.filer), return the final portion (e.g., filer).
     *
     * @param packageName Full package name.
     * @return Final portion of the name.
     */
    protected String finalPackageName(String packageName) {
        return packageName.substring(packageName.lastIndexOf('.') + 1);
    }
}
